import datetime

class Shared:

    @classmethod
    def init(cls, api_url, app_key: str, app_secret: str, private_key: str,
             institution: str, customer_number: str, user: str, password: str, user_id: str):
        cls._api_url = api_url
        cls._app_key = app_key
        cls._app_secret = app_secret
        cls._private_key = private_key
        cls._institution = institution
        cls._customer_number = customer_number
        cls._customer_token = None
        cls._server_token = None
        cls._user = user
        cls._password = password
        cls.user_id = user_id
        cls._ready = True

    @classmethod
    def destroy(cls):
        cls._ready = False
        cls._api_url = None
        cls._app_key = None
        cls._app_secret = None
        cls._private_key = None
        cls._institution = None
        cls._customer_number = None
        cls._customer_token = None
        cls._server_token = None

    @classmethod
    def set_assertion(cls, assertion):
        cls._assertion = assertion

    @classmethod
    def get_assertion(cls):
        return cls._assertion

    @classmethod
    def set_server_token(cls, server_token):
        cls._server_token = server_token

    @classmethod
    def get_server_token(cls):
        return cls._server_token

    @classmethod
    def get_server_token_expire_time(cls):
        """
        :return: the server token expire time in seconds
        """
        token = cls.get_server_token()
        # expire time in seconds
        expire_time = datetime.datetime.fromtimestamp(int(token['refreshTokenExpiresAt']) / 1000)
        return expire_time

    @classmethod
    def set_customer_token(cls, customer_token):
        cls._customer_token = customer_token

    @classmethod
    def get_customer_token(cls):
        return cls._customer_token

    @classmethod
    def get_customer_token_expire_time(cls):
        """
        :return: the server token expire time in seconds
        """
        token = cls.get_customer_token()
        # expire time in seconds
        expire_time = datetime.datetime.fromtimestamp(int(token['refreshTokenExpiresAt']) / 1000)
        return expire_time

    @classmethod
    def get_api_url(cls):
        return cls._api_url

    @classmethod
    def get_app_key(cls):
        return cls._app_key

    @classmethod
    def get_app_secret(cls):
        return cls._app_secret

    @classmethod
    def get_private_key(cls):
        return cls._private_key

    @classmethod
    def get_institution(cls):
        return cls._institution

    @classmethod
    def print(cls):
        print('App Key: ', cls._app_key)

    @classmethod
    def set_customer_number(cls, customer_number):
        cls._customer_number = customer_number

    @classmethod
    def get_customer_number(cls):
        return cls._customer_number

    @classmethod
    def get_user_name(cls):
        return cls._user

    @classmethod
    def get_password(cls):
        return cls._password

    @classmethod
    def get_user_id(cls):
        return cls.user_id



import json
import threading
from typing import Union
import gtnapi

import websocket


class Streaming:
    class MarketData:

        __active: bool

        @classmethod
        def __open_connection(cls, on_open, on_message, on_error, on_close):
            """
            open the websocket connection
            :param on_open: method reference
            :param on_message: method reference
            :param on_error: method reference
            :param on_close: method reference
            :return:
            """
            cls.__on_open = on_open
            cls.__on_message = on_message
            cls.__on_error = on_error
            cls.__on_close = on_close
            websocket.enableTrace(False)
            cls.__ws = websocket.WebSocketApp('wss' + gtnapi.get_api_url()[5:] + "/market-data/websocket/price?" + "throttle-key=" + gtnapi.get_app_key(),
                                              on_open=cls.on_open,
                                              on_message=cls.on_message,
                                              on_error=cls.on_error,
                                              on_close=cls.on_close)

            cls.__active = True
            cls.__ws.run_forever()
            cls.__active = False

        @classmethod
        def connect(cls, on_open, on_message, on_error, on_close):
            """
            Start the websocket thred
            :param on_open: method reference
            :param on_message: method reference
            :param on_error: method reference
            :param on_close: method reference
            :return:
            """
            t = threading.Thread(target=cls.__open_connection, args=(on_open, on_message, on_error, on_close))
            t.start()

        @classmethod
        def disconnect(cls):
            """
            disconnect the socket on demand
            :return:
            """
            cls.__active = False
            cls.__ws.close()

        @classmethod
        def on_open(cls, ws):
            """
            on open event handler
            :param ws: web socket reference
            """
            ws.send(f'{{ "token": {gtnapi.get_token()["accessToken"]} }}')
            cls.__on_open()

        @classmethod
        def on_message(cls, ws, message):
            """
            on message event handler
            :param ws: web socket reference
            :param message: message received
            """
            cls.__on_message(message)

        @classmethod
        def on_error(cls, ws, error):
            """
            on open error handler
            :param ws: web socket reference
            :param error: error message
            """
            cls.__on_error(error)

        @classmethod
        def on_close(cls, ws, close_status_code, close_msg):
            """
            on close event handler
            :param ws: web socket reference
            :param close_status_code: code
            :param close_msg: closing message
            """
            cls.__active = False
            cls.__on_close(close_status_code, close_msg)

        @classmethod
        def send_message(cls, message: Union[dict, str]):
            """
            sends a message via the web socket
            :param message: to be sent
            """
            if type(message) is dict:
                cls.__ws.send_text(json.dumps(message))
            else:
                cls.__ws.send_text(message)

        @classmethod
        def active(cls):
            """
            :return: True if market data streaming is active
            """
            try:
                return cls.__active
            except Exception as e:
                return False

    class TadeData:
        __active: bool

        @classmethod
        def __open_connection(cls, on_open, on_message, on_error, on_close):
            """
            open the websocket connection
            :param on_open: method reference
            :param on_message: method reference
            :param on_error: method reference
            :param on_close: method reference
            :return:
            """
            cls.__on_open = on_open
            cls.__on_message = on_message
            cls.__on_error = on_error
            cls.__on_close = on_close
            websocket.enableTrace(False)
            cls.__ws = websocket.WebSocketApp('wss' + gtnapi.get_api_url()[5:] + "/trade/websocket/v1.2.1?" + "throttle-key=" + gtnapi.get_app_key(),
                                              on_open=cls.on_open,
                                              on_message=cls.on_message,
                                              on_error=cls.on_error,
                                              on_close=cls.on_close)

            cls.__active = True
            cls.__ws.run_forever()
            cls.__active = False

        @classmethod
        def connect(cls, on_open, on_message, on_error, on_close):
            """
            Start the websocket thred
            :param on_open: method reference
            :param on_message: method reference
            :param on_error: method reference
            :param on_close: method reference
            :return:
            """
            t = threading.Thread(target=cls.__open_connection, args=(on_open, on_message, on_error, on_close))
            t.start()

        @classmethod
        def disconnect(cls):
            """
            disconnect the socket on demand
            :return:
            """
            cls.__active = False
            cls.__ws.close()

        @classmethod
        def on_open(cls, ws):
            """
            on open event handler
            :param ws: web socket reference
            """
            ws.send(f'{{ "token": {gtnapi.get_token()["accessToken"]} }}')
            cls.__on_open()

        @classmethod
        def on_message(cls, ws, message):
            """
            on message event handler
            :param ws: web socket reference
            :param message: message received
            """
            cls.__on_message(message)

        @classmethod
        def on_error(cls, ws, error):
            """
            on open error handler
            :param ws: web socket reference
            :param error: error message
            """
            cls.__on_error(error)

        @classmethod
        def on_close(cls, ws, close_status_code, close_msg):
            """
            on close event handler
            :param ws: web socket reference
            :param close_status_code: code
            :param close_msg: closing message
            """
            cls.__active = False
            cls.__on_close(close_status_code, close_msg)

        @classmethod
        def send_message(cls, message: Union[dict, str]):
            """
            sends a message via the web socket
            :param message: to be sent
            """
            if type(message) is dict:
                cls.__ws.send_text(json.dumps(message))
            else:
                cls.__ws.send_text(message)

        @classmethod
        def active(cls):
            """
            :return: True if market data streaming is active
            """
            try:
                return cls.__active
            except Exception as e:
                return False



from typing import Union, Callable

from gtnapi._shared import Shared as Shared
from gtnapi._auth import Auth as Auth
from gtnapi._streaming import Streaming as Streaming
from gtnapi._requests import Requests as Requests

shared: Shared


def init(api_url: str, institution: str, app_key: str,
         app_secret: str = None, private_key: str = None,
         customer_number: str = None,
         user: str = None, password: str = None, user_id: str = '1234'):
    """
    :param api_url: url of the api service
    :param app_key: app key of the institution
    :param app_secret: app secret of the institution
    :param private_key: private key of the institution
    :param institution: institution code
    :param customer_number: optional customer number. not required for the institution mode
    :param user: user name in user/pass mode
    :param user_id: user id tome associated with the institution / customer tokens
    :param password: password name in user/pass mode
    """
    global shared
    shared = Shared
    shared.init(api_url, app_key, app_secret, private_key, institution, customer_number, user, password, user_id)
    # key_rot._init()
    return Auth.init()


def get_api_url():
    """
    :return: the api service url
    """
    return shared.get_api_url()


def get_app_key():
    """
    :return: the app key
    """
    return shared.get_app_key()


def get_app_secret():
    """
    :return: the app secret
    """
    return shared.get_app_secret()


def get_private_key():
    """
    :return: the institution's private key
    """
    return shared.get_private_key()


def get_institution():
    """
    :return: the institution code
    """
    return shared.get_institution()


def set_assertion(assertion):
    """
    :param assertion: assertion string
    """
    shared.set_assertion(assertion)


def get_assertion():
    """
    :return: the assertion
    """
    return shared.get_assertion()


def set_server_token(server_token):
    """
    :param server_token: server token string
    """
    shared.set_server_token(server_token)


def get_server_token():
    """
    :return: the server token string
    """
    return shared.get_server_token()


def set_customer_token(customer):
    """
    :param customer: customer token string
    """
    shared.set_customer_token(customer)


def get_customer_token():
    """
    :return: the customer token
    """
    return shared.get_customer_token()


def get_user_name():
    """
    :return: the user name
    """
    return shared.get_user_name()


def get_password():
    """
    :return: the password
    """
    return shared.get_password()


def get_user_id():
    """
    :return: the password
    """
    return shared.get_user_id()

@property
def access_token():
    if is_customer_mode():
        return get_customer_token()
    else:
        return get_server_token()


def get_token():
    """
    :return: customer token if in customer mode or server token otherwise
    """
    if is_customer_mode():
        return get_customer_token()
    else:
        return get_server_token()


def set_customer_number(customer_number):
    """
    :param customer_number: set the customer number
    """
    shared.set_customer_number(customer_number)


def get_customer_number():
    """
    :return: the customer number
    """
    if is_user_mode():
        return get_customer_token()['customerNumber']
    else:
        return shared.get_customer_number()


def is_customer_mode():
    """
    :return: True if in customer mode
    """
    return shared.get_customer_number() is not None or shared.get_user_name() is not None


def is_server_mode():
    """
    :return: True if in server mode
    """
    return shared.get_customer_number() is None and shared.get_user_name() is None


def is_user_mode():
    """
    :return: True if in server mode
    """
    return shared.get_user_name() is not None


def stop():
    """
    stop and shutdown the connection.
    APIs are not accessible after this
    """
    Auth._shut_down()
    if Streaming.MarketData.active():
        Streaming.MarketData.disconnect()



import json

import requests
from urllib.parse import urlencode


class Connection:
    """
    Connection object to keep the http connection alive
    for the performance
    """

    def __init__(self, url, auth_token, throttle_key):
        self.url = url

        self.http_session = requests.Session()
        self.http_session.keep_alive = False
        if auth_token:
            self.http_session.headers.update(
                {'Authorization': auth_token, 'Throttle-Key': throttle_key, 'Content-Type': 'application/json'})
        else:
            self.http_session.headers.update(
                {'Throttle-Key': throttle_key, 'Content-Type': 'application/json'})

    def open_post(self, data):
        """
        Open the http connection
        """
        response = self.http_session.post(self.url, json.dumps(data))
        return response

    def open_get(self, data):
        """
        Open the http connection
        """
        response = self.http_session.get(self.url, params=data)
        return response

    def open_patch(self, data):
        """
        Open the http connection
        """
        response = self.http_session.patch(self.url, json.dumps(data))
        return response

    def open_delete(self, data):
        """
        Open the http connection
        """
        response = self.http_session.delete(self.url, params=urlencode(data))
        return response

    def close(self):
        """
        close the http session
        """
        self.http_session.close()



import json

import gtnapi
from gtnapi._connection import Connection
from typing import TypedDict


class Requests:

    @classmethod
    def __send_request(cls, method, endpoint, params) -> TypedDict('response', {'http_status': int, 'response': dict}):
        """
        send a request to the backend
        :param method: HTTP method
        :param endpoint: endpoint to call
        :param params: parameters to tbe send to the endpoint
        :return: HTTP status and the response dict
        """
        con = None
        try:
            # make sure to begin with a '/'
            if not endpoint[0] == '/':
                endpoint = '/' + endpoint

            con = Connection(gtnapi.get_api_url() + endpoint,
                             'Bearer ' + gtnapi.get_token()['accessToken'],
                             gtnapi.get_app_key())
            if method == 'POST':
                r = con.open_post(params)
            elif method == 'GET':
                r = con.open_get(params)
            elif method == 'PATCH':
                r = con.open_patch(params)
            elif method == 'DELETE':
                r = con.open_delete(params)
            else:
                return {
                    "http_status": 405,
                    "response": {
                        "status": "FAILED",
                        "reason": "unknown method: " + method
                    }
                }

            if r.status_code == 200:
                return {
                    "http_status": r.status_code,
                    "response": json.loads(r.text)
                }
            else:
                try:
                    return {
                        "http_status": r.status_code,
                        "response": json.loads(r.text)
                    }
                except Exception as e:
                    # possible JSON decode error
                    return {
                        "http_status": r.status_code,
                        "response": {}
                    }
        except Exception as e:
            print(e)
            return {
                "http_status": -1,
                "response": {}
            }
        finally:
            if con:
                con.close()

    @classmethod
    def get(cls, endpoint: str, **kwargs) -> dict:
        """
        call to the HTTP GET method
        :param endpoint: endpoint to call
        :param kwargs: parameters to tbe send to the endpoint
        :return: HTTP status and the response dict
        """
        return cls.__send_request("GET", endpoint, kwargs)

    @classmethod
    def post(cls, endpoint: str, **kwargs) -> dict:
        """
        call to the HTTP POST method
        :param endpoint: endpoint to call
        :param kwargs: parameters to tbe send to the endpoint
        :return: HTTP status and the response dict
        """
        return cls.__send_request("POST", endpoint, kwargs)

    @classmethod
    def patch(cls, endpoint: str, **kwargs) -> dict:
        """
        call to the HTTP PATCH method
        :param endpoint: endpoint to call
        :param kwargs: parameters to tbe send to the endpoint
        :return: HTTP status and the response dict
        """
        return cls.__send_request("PATCH", endpoint, kwargs)

    @classmethod
    def delete(cls, endpoint: str, **kwargs) -> dict:
        """
        call to the HTTP DELETE method
        :param endpoint: endpoint to call
        :param kwargs: parameters to tbe send to the endpoint
        :return: HTTP status and the response dict
        """
        return cls.__send_request("DELETE", endpoint, kwargs)



from typing import Dict, Union, TypedDict

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
import jwt
import json
from enum import Enum

import base64
import datetime, time
from gtnapi._connection import Connection
import gtnapi
import threading
import logging
import hashlib
import binascii


class Auth:
    class Statuses(Enum):
        AUTH_SUCCESS = 'SUCCESS'
        AUTH_FAILED = 'FAILED'
        ASSERTION_ERROR = 'ASSERTION_ERROR'
        SEVER_AUTH_FAILED = 'SEVER_AUTH_FAILED'
        CUSTOMER_AUTH_FAILED = 'CUSTOMER_AUTH_FAILED'
        SERVER_TOKEN_RENEWED = 'SERVER_TOKEN_RENEWED'
        SERVER_TOKEN_RENEW_FAILED = 'SERVER_TOKEN_RENEW_FAILED'
        CUSTOMER_TOKEN_RENEWED = 'CUSTOMER_TOKEN_RENEWED'
        CUSTOMER_TOKEN_RENEW_FAILED = 'CUSTOMER_TOKEN_RENEWED_FAILED'
        AUTH_EXPIRED = 'TOKEN_EXPIRED'

    @classmethod
    def init(cls):
        if gtnapi.is_user_mode():
            return cls.init_user()
        else:
            return cls.init_institution()

    @classmethod
    def init_user(cls):
        """
        init in user/pass mode
        get the customer token only
        """

        cls.logger = logging.getLogger(__name__)

        token_status = gtnapi.Auth.Statuses.AUTH_SUCCESS.value

        status, customer_token = cls._get_customer_token_for_user(gtnapi.get_user_name(), gtnapi.get_password(),
                                                                  gtnapi.get_institution())

        if status == 200 and customer_token is not None and customer_token['status'] == 'SUCCESS':
            cls.logger.info('Customer authentication success')
            gtnapi.shared.set_customer_token(customer_token)
            cls.logger.info('GTN API initiated in Customer mode.')
        else:
            if customer_token is None:
                cls.logger.error(f"Customer authentication failed. http status {status} and token is None")
            else:
                cls.logger.error(
                    f"Customer authentication failed. status {status} and token status {customer_token['status']}")
            token_status = gtnapi.Auth.Statuses.CUSTOMER_AUTH_FAILED.value

        if token_status == gtnapi.Auth.Statuses.AUTH_SUCCESS.value:
            # start the token refresh thread
            cls._start_thread()
            pass
        else:
            if status == 200:
                status = -1

        return cls.__return(status, token_status)

    @classmethod
    def init_institution(cls):
        """
        Initialise the authentication system
            1. create the basic auth token
            2. create the assertion
            3. get the server token
            4. if customer mode, get the customer token
        """

        cls.logger = logging.getLogger(__name__)

        token_status = gtnapi.Auth.Statuses.AUTH_SUCCESS.value

        # create the basic auth
        cls._basic_auth = cls._get_basic_auth(gtnapi.get_app_key(), gtnapi.get_app_secret())

        # create the assertion
        _assertion = cls._create_token(gtnapi.get_private_key(), gtnapi.get_app_key(),
                                       gtnapi.get_institution(), gtnapi.get_user_id())
        if _assertion is None:
            cls.logger.error('Assertion created failed')
            return cls.__return(-1,  gtnapi.Auth.Statuses.ASSERTION_ERROR.value)

        gtnapi.set_assertion(_assertion)
        cls.logger.info('Assertion created successfully')

        # get the server token
        status, server_token = cls._get_server_token()
        if status == 200 and server_token is not None and server_token['status'] == 'SUCCESS':
            cls.logger.info('Institution authentication success')
            # server_token['accessTokenExpiresAt'] = int(
            #     (datetime.datetime.now() + datetime.timedelta(minutes=2)).timestamp() * 1000)
            gtnapi.set_server_token(server_token)
            # get the customer token
            if gtnapi.is_customer_mode():
                status, customer_token = cls._get_customer_token()
                if status == 200 and customer_token is not None and customer_token['status'] == 'SUCCESS':
                    cls.logger.info('Customer authentication success')
                    # customer_token['accessTokenExpiresAt'] = int(
                    #     (datetime.datetime.now() + datetime.timedelta(minutes=2)).timestamp() * 1000)
                    gtnapi.shared.set_customer_token(customer_token)
                    cls.logger.info('GTN API initiated in Customer mode.')
                else:
                    if customer_token is None:
                        cls.logger.error(f"Customer authentication failed. http status {status} and token is None")
                    else:
                        cls.logger.error(
                            f"Customer authentication failed. status {status} and token status {customer_token['status']}")
                    token_status = gtnapi.Auth.Statuses.CUSTOMER_AUTH_FAILED.value
            else:
                cls.logger.debug('GTN API initiated in Institution mode')
        else:
            if server_token is None:
                cls.logger.error(f"Institution authentication failed. http status {status} and token is None")
            else:
                cls.logger.error(
                    f"Institution authentication failed. http status {status} and token status {server_token['status']}")
            token_status = gtnapi.Auth.Statuses.SEVER_AUTH_FAILED.value

        if token_status == gtnapi.Auth.Statuses.AUTH_SUCCESS.value:
            # start the token refresh thread
            cls._start_thread()
        else:
            if status == 200:
                status = -1

        # return {'http_code': status, 'status': token_status}
        return cls.__return(status, token_status)

    @classmethod
    def __logout(cls):
        """
        logout the service
        """
        gtnapi.shared.destroy()

    @classmethod
    def __refresh(cls):
        """
        token refresh process
        works in a different thread
        """
        cls.logger.debug('Token refresh in progress ...')
        # if in server mode, refresh the server token
        if gtnapi.is_server_mode():
            status, server_token = cls._get_server_token_refresh()
            if status == 200:
                cls.logger.debug('Refresh server token success')
                gtnapi.set_server_token(server_token)
            else:
                cls.logger.debug('Refresh server token failed')
        else:
            # if in customer mode, refresh the server token
            status, customer_token = cls._get_customer_token_refresh()
            if status == 200:
                cls.logger.debug('Refresh customer token success')
                gtnapi.set_customer_token(customer_token)
            else:
                cls.logger.debug('Refresh customer token failed')

    @classmethod
    def _create_token(cls, private_key, app_key, institution, user_id) -> Union[Dict, None]:
        """
        Create the JWT token
        :param private_key: private key of the institute
        :param app_key: app key assigned to the institute
        :param institution: institution code
        :return: the JWT token
        """
        try:

            private_key_byte_array = cls._base16Decoder(private_key)

            # Create RSA private key object from raw bytes
            private_key = serialization.load_der_private_key(private_key_byte_array, password=None)

            issued_time = int(datetime.datetime.now().timestamp() * 1)
            expire_time = int((datetime.datetime.now() + datetime.timedelta(hours=1)).timestamp() * 1)

            if isinstance(private_key, rsa.RSAPrivateKey):
                access_token = jwt.encode(
                    {
                        "iss": app_key,
                        "instCode": institution,
                        "exp": expire_time,
                        "userId": user_id,
                        "iat": issued_time
                    },
                    private_key,
                    'RS256'
                )
                return access_token

        except Exception as e:
            cls.logger.exception('Error creating assertion')

        return None

    @classmethod
    def _get_server_token(cls) -> tuple[int, Union[Dict, None]]:
        """
        get the server token from the backend
        :return: the server token received
        """
        con = None
        try:
            con = Connection(gtnapi.get_api_url() + '/trade/auth/token', cls._basic_auth, gtnapi.get_app_key())
            data = {'assertion': gtnapi.get_assertion()}
            r = con.open_post(data=data)
            if r.status_code == 200:
                return r.status_code, json.loads(r.text)
            else:
                return r.status_code, None
        except Exception as e:
            cls.logger.exception(e)
            return -1, None
        finally:
            if con:
                con.close()

    @classmethod
    def _get_server_token_refresh(cls) -> tuple[int, Union[Dict, None]]:
        """
        refresh the server token
        :return: the new server token
        """
        con = None
        try:
            con = Connection(gtnapi.get_api_url() + '/trade/auth/token/refresh', None, gtnapi.get_app_key())
            data = {'refreshToken': gtnapi.get_server_token()['refreshToken']}
            r = con.open_post(data=data)
            if r.status_code == 200:
                # cls.__raise_event(gtnapi.Auth.Statuses.SERVER_TOKEN_RENEWED)
                return r.status_code, json.loads(r.text)
            else:
                # cls.__raise_event(gtnapi.Auth.Statuses.SERVER_TOKEN_RENEW_FAILED)
                return r.status_code, None
        except Exception as e:
            cls.logger.exception(e)
            # cls.__raise_event(gtnapi.Auth.Statuses.SERVER_TOKEN_RENEW_FAILED)
            return -1, None
        finally:
            if con:
                con.close()

    @classmethod
    def _get_customer_token(cls) -> tuple[int, Union[Dict, None]]:
        """
        get the customer token from the backend
        :return: the customer token received
        """
        con = None
        try:
            con = Connection(gtnapi.get_api_url() + '/trade/auth/customer/token', cls._basic_auth, gtnapi.get_app_key())
            data = {'customerNumber': gtnapi.get_customer_number(), 'accessToken': gtnapi.get_server_token()['accessToken']}
            r = con.open_post(data=data)
            if r.status_code == 200:
                cls.logger.debug(f"Customer Token:{r.text}")
                return r.status_code, json.loads(r.text)
            else:
                return r.status_code, None
        except Exception as e:
            cls.logger.exception(e)
            return -1, None
        finally:
            if con:
                con.close()

    @classmethod
    def _get_customer_token_refresh(cls) -> tuple[int, Union[Dict, None]]:
        """
        refresh the customer token
        :return: the new customer token
        """
        con = None
        try:
            con = Connection(gtnapi.get_api_url() + '/trade/auth/customer/token/refresh', None, gtnapi.get_app_key())
            data = {'refreshToken': gtnapi.get_customer_token()['refreshToken']}
            r = con.open_post(data=data)
            if r.status_code == 200:
                cls.logger.debug(f"New Customer Token: {r.text}")
                # cls.__raise_event(gtnapi.Auth.Statuses.CUSTOMER_TOKEN_RENEWED)
                return r.status_code, json.loads(r.text)
            else:
                # cls.__raise_event(gtnapi.Auth.Statuses.CUSTOMER_TOKEN_RENEW_FAILED)
                return r.status_code, None
        except Exception as e:
            cls.logger.exception(e)
            # cls.__raise_event(gtnapi.Auth.Statuses.CUSTOMER_TOKEN_RENEW_FAILED)
            return -1, None
        finally:
            if con:
                con.close()

    @classmethod
    def _get_customer_token_for_user(cls, user: str, passw: str, institution: str) -> tuple[int, Union[Dict, None]]:
        """
        get the customer token from the backend
        :return: the customer token received
        """
        con = None
        try:
            encoded_pw = cls.hash_password(passw)
            con = Connection(gtnapi.get_api_url() + '/trade/auth/user-login', None, gtnapi.get_app_key())
            data = {'loginName': user, 'password': encoded_pw, 'institutionCode': institution, 'encryptionType': 2}
            r = con.open_post(data=data)
            if r.status_code == 200:
                cls.logger.debug(f"Customer Token:{r.text}")
                return r.status_code, json.loads(r.text)
            else:
                return r.status_code, None
        except Exception as e:
            cls.logger.exception(e)
            return -1, None
        finally:
            if con:
                con.close()

    @classmethod
    def _base16Decoder(cls, hex_string):
        bts = bytearray(len(hex_string) // 2)
        for i in range(len(bts)):
            bts[i] = int(hex_string[2 * i:2 * i + 2], 16)
        return bytes(bts)

    @classmethod
    def _base16Encoder(cls, digestMsgByte):
        verifyMsg = ""
        for i in range(len(digestMsgByte)):
            hexChar = digestMsgByte[i] & 0xFF
            hexString = hex(hexChar)[2:]
            if len(hexString) == 1:
                verifyMsg += "0"
            verifyMsg += hexString
        return verifyMsg

    @classmethod
    def _get_basic_auth(cls, username, password) -> str:
        """
        create the basic auth token
        :param username: user name
        :param password: password
        :return:
        """
        token = base64.b64encode(f"{username}:{password}".encode('utf-8')).decode("ascii")
        return f'Basic {token}'

    @property
    def basic_auth(self):
        return self._basic_auth

    @classmethod
    def _key_refresh(cls):
        """
        token refresh thread
        refreshes the token within last 100 seconds of the token lifetime
        """
        while cls.__thread_active:
            try:
                if gtnapi.is_server_mode():
                    token = gtnapi.get_server_token()
                    # check the refresh token validity first
                    expire_time = datetime.datetime.fromtimestamp(int(token['refreshTokenExpiresAt']) / 1000)
                    delta = (expire_time - datetime.datetime.now()).total_seconds()
                    cls.logger.debug(f'Time delta - server refresh token:  {delta}s')
                    if delta < 50:
                        # refresh token no longer valid. must logout
                        cls._shut_down()
                        gtnapi.Auth.__logout()
                        # cls.__raise_event(gtnapi.Auth.Statuses.AUTH_EXPIRED)
                    else:
                        # refresh token valid lets check the access token
                        expire_time = datetime.datetime.fromtimestamp(int(token['accessTokenExpiresAt']) / 1000)
                        delta = (expire_time - datetime.datetime.now()).total_seconds()
                        cls.logger.debug(f'Time delta server access token: {delta}')
                        if 0 < delta < 100:
                            cls.logger.debug(f'refreshing institution access token')
                            status, server_token = cls._get_server_token_refresh()
                            if status == 200:
                                cls.logger.debug(f'refreshing institution access token successful')
                                # server_token['refreshTokenExpiresAt'] = int(
                                #     (datetime.datetime.now() + datetime.timedelta(minutes=2)).timestamp() * 1000)
                                gtnapi.shared.set_server_token(server_token)
                                cls._print_token()
                            else:
                                cls.logger.debug(f'Error refreshing the token: {delta}')
                else:
                    token = gtnapi.get_customer_token()
                    # check the refresh token validity first
                    expire_time = datetime.datetime.fromtimestamp(int(token['refreshTokenExpiresAt']) / 1000)
                    delta = (expire_time - datetime.datetime.now()).total_seconds()
                    cls.logger.debug(f'Time delta - customer refresh token: {delta}')
                    if delta < 50:
                        # refresh token no longer valid. must logout
                        cls._shut_down()
                        gtnapi.Auth.__logout()
                        # cls.__raise_event(gtnapi.Auth.Statuses.AUTH_EXPIRED)
                    else:
                        # refresh token valid lets check the access token
                        expire_time = datetime.datetime.fromtimestamp(int(token['accessTokenExpiresAt']) / 1000)
                        delta = (expire_time - datetime.datetime.now()).total_seconds()
                        cls.logger.debug(f'Time delta customer access token: {delta}')
                        if 0 < delta < 100:
                            cls.logger.debug(f'refreshing customer access token')
                            status, customer_token = cls._get_customer_token_refresh()
                            if status == 200:
                                cls.logger.debug(f'refreshing customer access token successful')
                                # customer_token['refreshTokenExpiresAt'] = int(
                                #     (datetime.datetime.now() + datetime.timedelta(minutes=2)).timestamp() * 1000)
                                gtnapi.shared.set_customer_token(customer_token)
                                cls._print_token()
                            else:
                                cls.logger.debug(f'Error refreshing the token: {status}')
            except Exception as e:
                cls.logger.exception(e)

            time.sleep(10)

        cls.logger.debug('Key rotation thread exit!')

    @classmethod
    def _start_thread(cls):
        """
        start the token refresh thread
        """
        cls.logger.debug('Starting refresh thread')
        cls.__thread_active = True
        cls.t = threading.Thread(target=cls._key_refresh)
        cls.t.start()

    @classmethod
    def _print_token(cls):
        if gtnapi.is_server_mode():
            token = gtnapi.get_server_token()
        else:
            token = gtnapi.get_customer_token()
        cls.logger.debug(json.dumps(token, indent=4))
        cls.logger.debug('Access', token['accessToken'])
        cls.logger.debug("Access expire", datetime.datetime.fromtimestamp(int(token['accessTokenExpiresAt']) / 1000))
        cls.logger.debug('Refresh', token['refreshToken'])
        cls.logger.debug("Refresh expire", datetime.datetime.fromtimestamp(int(token['refreshTokenExpiresAt']) / 1000))

    @classmethod
    def _shut_down(cls):
        """
        stop the refresh thread
        """
        cls.__logout()
        cls.__thread_active = False
        cls.logger.info('GTN API session terminated')

    @classmethod
    def hash_password(cls, password: str) -> Union[str, None]:
        salt = b"MUBASHER"  # Convert the salt to bytes
        iterations = 10000
        key_length = 64  # 512 bits / 8 = 64 bytes

        try:
            # Derive the key using PBKDF2 with HMAC-SHA-512
            hash_bytes = hashlib.pbkdf2_hmac('sha512', password.encode(), salt, iterations, key_length)

            # Convert the resulting byte array into a hexadecimal string
            return binascii.hexlify(hash_bytes).decode()
        except Exception as e:
            cls.logger.exception(f"Error obtaining hashPassword: {e}")
            return None

    @classmethod
    def __return(cls, status: int, response: Union[dict, str]) -> TypedDict('response', {'http_status': int, 'auth_status': Union[str, dict]}):
        return {
            "http_status": status,
            "auth_status": response
        }
